package scalaCollectionNull

class SinglyLinkedList[T] extends MyList[T] {
	/**
	 * The number of elements in list. Initially empty
	 */
	protected var count: Int = 0
 

	/**
	 * The head of the list. An (optional) reference to a singly linked list node.
	 * Value is None if the list is empty
	 */
	protected var head: Option[MyNode[T]] = None

	/**
	 * Add a value to head of list.
	 * 
	 * @post value is added to beginning of list
	 * 
	 * @param value
	 *            The value to be added to head of list.
	 * overridden here to be more efficient 
	 */
	override def addFirst(value: T) = {
		head = Some(new MyNode(value,head));
		count+=1;
	}
 

	/**
	 * Add a value to head of list.
	 * 
	 * @post value is added to beginning of list
	 * 
	 * @param value
	 *            The value to be added to head of list.
	 * overridden here to be more efficient 
	 */
	override def addLast(value: T) = add(size,value)
  
	/**
	 * Remove first element of list.
	 * 
	 * @pre list is not empty
	 * @post removes and returns node from beginning of list
	 * 
	 * @return The value actually removed.
	 * overridden here to be more efficient  
	 */
	override def removeFirst(): T = {
	  	head match {
	  	  case Some(hdr) => {
	  		head = hdr.next; // move head down list
	  		count -= 1;
	  		hdr.elt
	  	  } 
	  	  case None => throw new Exception("Can't pop from empty list")
	  	}
	}


	/**
	 * Remove a node from list. 
	 * 
	 * @pre node is not null
	 * @post node has been removed
	 * 
	 * @param node
	 *            The node to be removed.
	 */
	def remove(node: MyNode[T]) = {
	  head match {
	    case None => throw new IllegalArgumentException("argument node not in list")
	    case Some(hd) => {
	    	if (hd.equals(node)) {
	    		head = hd.next
	    	} else {
	    		def testNextNd(nd:MyNode[T]): Boolean =  nd.next.equals(Some(node))
	    		var previous = hd.findProp(testNextNd)
	    		previous match {
	    			case None => throw new IllegalArgumentException("argument node not in list")
	    			case Some(nprev) => {
	    				nprev.next = node.next;
	    			}
	    		} 
	    	} 
		}
  		count -= 1;
	  }
	}

	/**
	 * Determine number of elements in list.
	 * 
	 * @post list is unchanged
	 * @return The number of elements in list.
	 */
	def size(): Int = count

	/**
	 * Remove all values from list.
	 * 
	 * @post all elements have been removed from list
	 */
	def clear() = {
		head = None;
		count = 0;
	}

	/**
	 * Get node at location i.
	 * 
	 * @pre 0 <= i < size()
	 * @post list is unchanged
	 * 
	 * @param i
	 *            position of value to be retrieved.
	 * @return value retrieved from location i 
	 */
	def get(i: Int): T = {
	  	getNode(i) match {
	  	  case None => throw new NoSuchElementException("No ith element")
	  	  case Some(nd) => nd.elt
	  	}
	} 
 
 	def getNode(i: Int): Option[MyNode[T]] = {
		head match {
			case Some(hd) => hd.getIth(i)
			case None => None
		}
	}


	/**
	 * Set value stored at location i to object o, returning old value.
	 * 
	 * @pre 0 <= i < size()
	 * @post ith entry of list has been set to value o
	 * @param i
	 *            location of entry to be changed.
	 * @param o
	 *            new value
	 * @return former value of ith entry of list.
	 */
	def set(i: Int, newValue: T): T = {
		head match {
		  case None => throw new IllegalArgumentException("no "+i+"th argument in list")
		  case Some(hdr) => {
			  hdr.getIth(i) match {
			    case None => throw new IllegalArgumentException("no "+i+"th argument in list")
			    case Some(updateNode) => {
			    	val oldVal = updateNode.elt
			    	updateNode.elt = newValue
			    	oldVal
			    } 
			  } 
		  } 
		} 
	} 

	/**
	 * Insert value at location.
	 * 
	 * @pre 0 <= i <= size()
	 * @post added value to be ith entry of list
	 * @param i
	 *            index of this new value
	 * @param value
	 *            value to be stored
	 */
	def add(i: Int, value: T) = {
		if (i == 0) {
			addFirst(value);
		} else {
			getNode(i-1) match {
			  case None => throw new NoSuchElementException("attempt to add past end of list")
			  case Some(prev) => {
				  val newNode = new MyNode(value, prev.next)
				  count += 1;
				  // make previous node point to new node
				  prev.next = Some(newNode);
			  } 
			} 
		}
	}

	/**
	 * Remove and return value at location i.
	 * 
	 * @pre 0 <= i < size()
	 * @post removes and returns object found at that location i
	 * 
	 * @param i
	 *            position of value to be retrieved.
	 * @return value retrieved from location i (returns null if i invalid)
	 */
	def removeIth(i: Int): T = {
		if (i == 0) {
			removeFirst();
		} else {
			getNode(i-1) match {
			  case None => throw new IllegalArgumentException("no "+i+"th argument in list")
			  case Some(prev) => {
				prev.next match {
					case Some(removed) => {
						// create new value to insert in correct position
						count -= 1;
						// make previous value point to new value
						prev.next = removed.next;
						removed.elt
					}
					case None => throw new IllegalArgumentException("no "+i+"th argument in list")
				} 
			  }  
			} 
		} 
	}

	/**
	 * Determine first location of a value in list.
	 * 
	 * @pre value is not null
	 * @post returns the (0-origin) index of value, or -1 if value is not found
	 * 
	 * @param value
	 *            value sought
	 * @return index (0 is first element) of value, or -1
	 */
	def indexOf(value: T): Int = {
		head match {
		  case Some(hdr) => hdr.indexOf(value)
		  case None => -1
		} 
	}

	/**
	 * Returns an iterator traversing list from head to tail.
	 * 
	 * @post returns enumeration allowing traversal of list
	 * 
	 * @return An iterator to traverse list.
	 */

	def iterator(): Iterator[T] = {
		new LinkedListIterator(head);
	}

	/**
	 * Construct a string representing list.
	 * 
	 * @post returns a string representing list
	 * 
	 * @return A string representing list.
	 */
	override def toString(): String = {
		"LinkedList:\n"+super.toString();
	}
	
    /**
     * Return a node with value val.
     *
     * @return returns a node with holding value val
     */
    def makeNode(v: T): MyNode[T] = {
    	return new MyNode[T](v);
    }


	private class LinkedListIterator[T](var current: Option[MyNode[T]]) extends Iterator[T] {		

		var currentIndex: Int = 0

		/**
		 * Determine if the iteration is finished.
		 * 
		 * @post returns true if there is more structure to be viewed: i.e., if
		 *       value (next) can return a useful value.
		 * 
		 * @return True if the iterator has more elements to be considered.
		 */
		def hasNext: Boolean = {
			current != None;
		}

		/**
		 * Return current value and increment Iterator.
		 * 
		 * @pre traversal has more elements
		 * @post returns current value and increments iterator
		 * 
		 * @return The current value, before increment.
		 */
		def next: T = {
			current match {
			  case Some(ncurrent) => {
				  val value = ncurrent.elt
				  current = ncurrent.next
			  	  currentIndex += 1;
				  return value
			  }
			  case None => throw new Exception("No elements left in iterator")
			}

		}

		/**
		 * Return structure's current object reference.
		 * 
		 * @pre traversal has more elements
		 * @post returns current value referenced by iterator
		 * 
		 * @return Object currently referenced.
		 */
		def get: T = {
			current match {
			  case Some(ncurrent) => ncurrent.elt
			  case None => throw new Exception("No elements left in iterator")
			}
		 }


		/**
		 * Returns the index of the element that would be returned by a
		 * subsequent call to next.
		 * 
		 * @post returns index of next element if there is more structure to be
		 *       viewed: and length of list otherwise
		 */
		def nextIndex: Int = currentIndex
  
		override def toString = "Iterator"+current

	}
}

object TestSinglyLinkedList {
  def main(args: Array[String]){
	  var test1 = new SinglyLinkedList[String]
	  test1.add(0,"abc")
	  println(test1.get(0)+" with count = "+test1.size)
	  test1.iterator.foreach(arg =>  println(arg))
	  println(test1.get(0)+" with count = "+test1.size)
	  test1.add(1, "d")
	  println(test1.get(1)+" with count = "+test1.size)
	  test1.add(2,"e")
	  println(test1.get(2)+" with count = "+test1.size)
	  println("Entire list: "+test1)
	  println("index of d "+test1.indexOf("d"))
	  println("Test remove")
	  val removed = test1.removeIth(0)
	  println("Drop "+removed+" to get "+test1)
   
	  for (i <- 0 to test1.size-1) {
	    println(test1(i))
	  }
   }
}

